@@ -1,12 +1,15 @@
 local bitshim = require("bitshim")
+local luautf8 = require("lua-utf8")
+local semver = require("semver")
+
 local utilities = {}
 
 local epsilon = 1E-12
 
-utilities.required = function (options, name, context, _type)
+utilities.required = function (options, name, context, required_type)
   if not options[name] then utilities.error(context.." needs a "..name.." parameter") end
-  if _type then
-    return utilities.cast(_type, options[name])
+  if required_type then
+    return utilities.cast(required_type, options[name])
   end
   return options[name]
 end
@@ -22,21 +25,31 @@ utilities.boolean = function (value, default)
   if value == "true" then return true end
   if value == "no" then preferbool(); return false end
   if value == "yes" then preferbool(); return true end
+  if value == nil then return default end
+  SU.error("Expecting a boolean value but got '" .. value .. "'")
   return default
 end
 
+local _skip_traceback_levels = 2
+
 utilities.error = function(message, bug)
+  _skip_traceback_levels = 3
   utilities.warn(message, bug)
+  _skip_traceback_levels = 2
   io.stderr:flush()
-  SILE.outputter:finish()
-  os.exit(1)
+  SILE.outputter:finish() -- Only really useful from the REPL but no harm in trying
+  SILE.scratch.caughterror = true
+  error(message, 2)
 end
 
 utilities.warn = function(message, bug)
+  if SILE.quiet then return end
   io.stderr:write("\n! " .. message)
   if SILE.traceback or bug then
     io.stderr:write(" at:\n" .. SILE.traceStack:locationTrace())
-    io.stderr:write(debug.traceback(nil, 2) or "\t! debug.traceback() did not identify code location")
+    if _skip_traceback_levels == 2 then
+      io.stderr:write(debug.traceback("", _skip_traceback_levels) or "\t! debug.traceback() did not identify code location")
+    end
   else
     io.stderr:write(" at " .. SILE.traceStack:locationHead())
   end
@@ -44,7 +57,7 @@ utilities.warn = function(message, bug)
 end
 
 utilities.debugging = function (category)
-  return SILE.debugFlags.all or SILE.debugFlags[category]
+  return SILE.debugFlags.all and category ~= "profile" or SILE.debugFlags[category]
 end
 
 utilities.feq = function (lhs, rhs) -- Float point equal
@@ -77,32 +90,77 @@ utilities.gtoke = function (string, pattern)
 end
 
 utilities.deprecated = function (old, new, warnat, errorat, extra)
+  warnat, errorat = semver(warnat or 0), semver(errorat or 0)
+  local current = SILE.version and semver(SILE.version:match("v([0-9]*.[0-9]*.[0-9]*)")) or warnat
   -- SILE.version is defined *after* most of SILE loads. It’s available at
   -- runtime but not useful if we encounter deprecated code in core code. Users
   -- will never encounter this failure, but as a developer it’s hard to test a
   -- deprecation when core code refactoring is an all-or-nothing proposition.
   -- Hence we fake it ‘till we make it, all deprecations internally are warings.
-  local _semver = SILE.version and SILE.version:match("v([0-9]*.[0-9]*.[0-9]*)") or warnat
-  local msg = old .. "() was deprecated in SILE v" .. warnat .. ". Please use " .. new .. "() instead. " .. extra
-  if errorat and _semver >= errorat then
+  local brackets = old:sub(1,1) == '\\' and "" or "()"
+  local _new = new and "Please use " .. (new .. brackets) .. " instead." or "Plase don't use it."
+  local msg = (old .. brackets) .. " was deprecated in SILE v" .. tostring(warnat) .. ". " .. _new ..  (extra and "\n" .. extra .. "\n\n" or "")
+  if errorat and current >= errorat then
     SU.error(msg)
-  elseif warnat and _semver >= warnat then
+  elseif warnat and current >= warnat then
     SU.warn(msg)
   end
 end
 
 utilities.debug = function (category, ...)
+  if SILE.quiet then return end
   if utilities.debugging(category) then
     local inputs = table.pack(...)
     for i, input in ipairs(inputs) do
       if type(input) == "function" then
-        inputs[i] = input()
+        local status, output = pcall(input)
+        inputs[i] = status and output or SU.warn(("Output of %s debug function was an error: %s"):format(category, output))
       end
     end
-    io.stderr:write("\n["..category.."] ", utilities.concat(inputs, " "))
+    local message = utilities.concat(inputs, " ")
+    if message then io.stderr:write(("\n[%s] %s"):format(category, message)) end
   end
 end
 
+utilities.debugAST = function (ast, level)
+  if not ast then
+    SU.error("debugAST called with nil", true)
+  end
+  local out = string.rep("  ", 1+level)
+  if level == 0 then
+    SU.debug("ast", function ()
+      return "[" .. SILE.currentlyProcessingFile
+    end)
+  end
+  if type(ast) == "function" then
+    SU.debug("ast", function ()
+      return out .. tostring(ast)
+    end)
+  elseif type(ast) == "table" then
+    for _, content in ipairs(ast) do
+      if type(content) == "string" then
+        SU.debug("ast", function ()
+          return out .. "[" .. content .. "]"
+        end)
+      elseif type(content) == "table" then
+        if SILE.Commands[content.command] then
+          SU.debug("ast", function ()
+            return out .. "\\" .. content.command .. " " .. pl.pretty.write(content.options, "")
+          end)
+          if (#content>=1) then utilities.debugAST(content, level+1) end
+        elseif content.id == "texlike_stuff" or (not content.command and not content.id) then
+          utilities.debugAST(content, level+1)
+        else
+          SU.debug("ast", function ()
+            return out .. "?\\" .. (content.command or content.id)
+          end)
+        end
+      end
+    end
+  end
+  if level == 0 then SU.debug("ast", "]") end
+end
+
 utilities.dump = function (...)
   local arg = { ... } -- Avoid things that Lua stuffs in arg like args to self()
   pl.pretty.dump(#arg == 1 and arg[1] or arg, "/dev/stderr")
@@ -247,11 +305,6 @@ utilities.cast = function (wantedType, value)
   wantedType = string.lower(wantedType)
   if wantedType:match(actualType)     then return value
   elseif actualType == "nil" and wantedType:match("nil") then return nil
-  elseif wantedType:match("integer") or wantedType:match("number") then
-    if type(value) == "table" and type(value.tonumber) == "function" then
-      return value:tonumber()
-    end
-    return tonumber(value)
   elseif wantedType:match("length")      then return SILE.length(value)
   elseif wantedType:match("measurement") then return SILE.measurement(value)
   elseif wantedType:match("vglue")       then return SILE.nodefactory.vglue(value)
@@ -260,16 +313,48 @@ utilities.cast = function (wantedType, value)
   elseif actualType == "nil" then SU.error("Cannot cast nil to " .. wantedType)
   elseif wantedType:match("boolean")     then return SU.boolean(value)
   elseif wantedType:match("string")      then return tostring(value)
+  elseif wantedType:match("number") then
+    if type(value) == "table" and type(value.tonumber) == "function" then
+      return value:tonumber()
+    end
+    local num = tonumber(value)
+    if not num then SU.error("Cannot cast '" .. value .. "'' to " .. wantedType) end
+    return num
+  elseif wantedType:match("integer") then
+    local num
+    if type(value) == "table" and type(value.tonumber) == "function" then
+      num = value:tonumber()
+    else
+      num = tonumber(value)
+    end
+    if not num then SU.error("Cannot cast '" .. value .. "'' to " .. wantedType) end
+    if not wantedType:match("number") and num % 1 ~= 0 then
+      -- Could be an error but since it wasn't checked before, let's just warn:
+      -- Some packages might have wrongly typed settings, for instance.
+      SU.warn("Casting an integer but got a float number " .. num)
+    end
+    return num
   else SU.error("Cannot cast to unrecognized type " .. wantedType)
   end
 end
 
+utilities.hasContent = function(content)
+  return type(content) == "function" or type(content) == "table" and #content > 0
+end
+
 -- Flatten content trees into just the string components (allows passing
 -- objects with complex structures to functions that need plain strings)
 utilities.contentToString = function (content)
   local string = ""
   for i = 1, #content do
-    if type(content[i]) == "string" then
+    if type(content[i]) == "table" and type(content[i][1]) == "string" then
+      string = string .. content[i][1]
+    elseif type(content[i]) == "string" then
+      -- Work around PEG parser returning env tags as content
+      -- TODO: refactor capture groups in PEG parser
+      if content.command == content[i] and content[i] == content[i+1] then
+        break
+      end
       string = string .. content[i]
     end
   end
@@ -300,6 +385,29 @@ utilities.walkContent = function (content, action)
   end
 end
 
+--- Strip position, line and column recursively from a content tree.
+-- This can be used to remove position details where we do not want them,
+-- e.g. in table of contents entries (referring to the original content,
+-- regardless where it was exactly, for the purpose of checking whether
+-- the table of contents changed.)
+--
+utilities.stripContentPos = function (content)
+  if type(content) ~= "table" then
+    return content
+  end
+  local stripped = {}
+  for k, v in pairs(content) do
+    if type(v) == "table" then
+      v = SU.stripContentPos(v)
+    end
+    stripped[k] = v
+  end
+  if content.id or content.command then
+    stripped.pos, stripped.col, stripped.lno = nil, nil, nil
+  end
+  return stripped
+end
+
 utilities.rateBadness = function(inf_bad, shortfall, spring)
   if spring == 0 then return inf_bad end
   local bad = math.floor(100 * math.abs(shortfall / spring) ^ 3)
@@ -368,9 +476,9 @@ utilities.utf16codes = function (ustr, endian)
       return nil
     else
       local c1, c2, c3, c4, wchar, lowchar
-      c1 = string.byte(ustr, pos)
+      c1 = string.byte(ustr, pos, pos+1)
       pos = pos + 1
-      c2 = string.byte(ustr, pos)
+      c2 = string.byte(ustr, pos, pos+1)
       pos = pos + 1
       if endian == "be" then
         wchar = c1 * 256 + c2
@@ -380,9 +488,9 @@ utilities.utf16codes = function (ustr, endian)
       if not (wchar >= 0xD800 and wchar <= 0xDBFF) then
         return wchar
       end
-      c3 = string.byte(ustr, pos)
+      c3 = string.byte(ustr, pos, pos+1)
       pos = pos + 1
-      c4 = string.byte(ustr, pos)
+      c4 = string.byte(ustr, pos, pos+1)
       pos = pos + 1
       if endian == "be" then
         lowchar = c3 * 256 + c4
@@ -395,29 +503,10 @@ utilities.utf16codes = function (ustr, endian)
 end
 
 utilities.splitUtf8 = function (str) -- Return an array of UTF8 strings each representing a Unicode char
-  local seq = 0
   local rv = {}
-  local val = -1
-  local this = ""
-  for i = 1, #str do
-    local c = string.byte(str, i)
-    if seq == 0 then
-      if val > -1 then
-        rv[1+#rv] = this
-        this = ""
-      end
-      seq = c < 0x80 and 1 or c < 0xE0 and 2 or c < 0xF0 and 3 or
-            c < 0xF8 and 4 or --c < 0xFC and 5 or c < 0xFE and 6 or
-          error("invalid UTF-8 character sequence")
-      val = bitshim.band(c, 2^(8-seq) - 1)
-      this = this .. str[i]
-    else
-      val = bitshim.bor(bitshim.lshift(val, 6), bitshim.band(c, 0x3F))
-      this = this .. str[i]
-    end
-    seq = seq - 1
+  for _, cp in luautf8.next, str do
+    table.insert(rv, luautf8.char(cp))
   end
-  rv[1+#rv] = this
   return rv
 end
 
@@ -431,121 +520,71 @@ utilities.firstChar = function (str)
   return chars[1]
 end
 
+local byte, floor, reverse = string.byte, math.floor, string.reverse
+
 utilities.utf8charat = function (str, index)
   return str:sub(index):match("([%z\1-\127\194-\244][\128-\191]*)")
 end
 
-utilities.utf8_to_utf16be_hexencoded = function (str)
-  local ustr = string.format("%04x", 0xfeff) -- BOM
-  for _, uchr in luautf8.codes(str) do
-    if (uchr < 0x10000) then
-      ustr = ustr..string.format("%04x", uchr)
-    else -- Surrogate pair
-      local sur_hi = (uchr - 0x10000) / 0x400 + 0xd800
-      local sur_lo = (uchr - 0x10000) % 0x400 + 0xdc00
-      ustr = ustr..string.format("%04x%04x", sur_hi, sur_lo)
-    end
+local utf16bom = function(endianness)
+  return endianness == "be" and "\254\255" or endianness == "le" and "\255\254" or SU.error("Unrecognized endianness")
+end
+
+utilities.hexencoded = function (str)
+  local ustr = ""
+  for i = 1, #str do
+    ustr = ustr..string.format("%02x", byte(str, i, i+1))
   end
   return ustr
 end
 
-utilities.utf8_to_utf16be = function (str)
+utilities.hexdecoded = function (str)
+  if #str % 2 == 1 then SU.error("Cannot decode hex string with odd len") end
   local ustr = ""
-  for _, uchr in luautf8.codes(str) do
-    if (uchr < 0x10000) then
-      ustr = ustr..string.format("%c%c", uchr / 256, uchr % 256 )
-    else -- Surrogate pair
-      local sur_hi = (uchr - 0x10000) / 0x400 + 0xd800
-      local sur_lo = (uchr - 0x10000) % 0x400 + 0xdc00
-      ustr = ustr..string.format("%c%c%c%c", sur_hi / 256, sur_hi % 256 , sur_lo / 256, sur_lo % 256)
-    end
+  for i = 1, #str, 2 do
+    ustr = ustr..string.char(tonumber(string.sub(str, i, i+1), 16))
   end
   return ustr
 end
 
-utilities.utf8_to_utf16le = function (str)
-  local ustr = ""
+local uchr_to_surrogate_pair = function(uchr, endianness)
+  local hi, lo = floor((uchr - 0x10000) / 0x400) + 0xd800, (uchr - 0x10000) % 0x400 + 0xdc00
+  local s_hi, s_lo = string.char(floor(hi / 256)) .. string.char(hi % 256), string.char(floor(lo / 256)) .. string.char(lo % 256)
+  return endianness == "le" and (reverse(s_hi) .. reverse(s_lo)) or s_hi .. s_lo
+end
+
+local uchr_to_utf16_double_byte = function(uchr, endianness)
+  local ustr = string.char(floor(uchr / 256)) .. string.char(uchr % 256)
+  return endianness == "le" and reverse(ustr) or ustr
+end
+
+local utf8_to_utf16 = function(str, endianness)
+  local ustr = utf16bom(endianness)
   for _, uchr in luautf8.codes(str) do
-    if (uchr < 0x10000) then
-      ustr = ustr..string.format("%c%c", uchr % 256, uchr / 256 )
-    else -- Surrogate pair
-      local sur_hi = (uchr - 0x10000) / 0x400 + 0xd800
-      local sur_lo = (uchr - 0x10000) % 0x400 + 0xdc00
-      ustr = ustr..string.format("%c%c%c%c", sur_hi % 256, sur_hi / 256 , sur_lo % 256, sur_lo / 256)
-    end
+    ustr = ustr..(uchr < 0x10000 and uchr_to_utf16_double_byte(uchr, endianness)
+                  or uchr_to_surrogate_pair(uchr, endianness))
   end
   return ustr
 end
 
-utilities.utf16le_to_utf8 = function (str)
+utilities.utf8_to_utf16be = function (str) return utf8_to_utf16(str, "be") end
+utilities.utf8_to_utf16le = function (str) return utf8_to_utf16(str, "le") end
+utilities.utf8_to_utf16be_hexencoded = function (str) return utilities.hexencoded(utilities.utf8_to_utf16be(str)) end
+utilities.utf8_to_utf16le_hexencoded = function (str) return utilities.hexencoded(utilities.utf8_to_utf16le(str)) end
+
+local utf16_to_utf8 = function (str, endianness)
+  local bom = utf16bom(endianness)
+
+  if str:find(bom) == 1 then str = string.sub(str, 3, #str) end
   local ustr = ""
-  for uchr in utilities.utf16codes(str, "le") do
+  for uchr in utilities.utf16codes(str, endianness) do
     ustr = ustr..luautf8.char(uchr)
   end
   return ustr
 end
 
-utilities.utf16be_to_utf8 = function (str)
-  local ustr = ""
-  for uchr in utilities.utf16codes(str, "be") do
-    ustr = ustr..luautf8.char(uchr)
-  end
-  return ustr
-end
-
-local icu = require("justenoughicu")
-
-local icuFormat = function (num, format)
-  local ok, result  = pcall(function() return icu.format_number(num, format) end)
-  return tostring(ok and result or num)
-end
-
--- Language specific number formatters add functions to this table,
--- see e.g. languages/tr.lua
-utilities.formatNumber = {
-  und = {
-
-    alpha = function (num)
-      local out = ""
-      local a = string.byte("a")
-      repeat
-        num = num - 1
-        out = string.char(num % 26 + a) .. out
-        num = (num - num % 26) / 26
-      until num < 1
-      return out
-    end
-
-  }
-}
-
-setmetatable (utilities.formatNumber, {
-    __call = function (self, num, format, case)
-      if math.abs(num) > 9223372036854775807 then
-        SU.warn("Integers larger than 64 bits do not reproduce properly in all formats")
-      end
-      if not case then
-        if format:match("^%l") then
-          case = "lower"
-        elseif format:match("^.%l") then
-          case = "title"
-        else
-          case = "upper"
-        end
-      end
-      local lang = SILE.settings.get("document.language")
-      format = format:lower()
-      local result
-      if self[lang] and type(self[lang][format]) == "function" then
-        result = self[lang][format](num)
-      elseif type(self["und"][format]) == "function" then
-        result = self.und[format](num)
-      else
-        result = icuFormat(num, format)
-      end
-      return icu.case(result, lang, case)
-    end
-})
+utilities.utf16be_to_utf8 = function (str) return utf16_to_utf8(str, "be") end
+utilities.utf16le_to_utf8 = function (str) return utf16_to_utf8(str, "le") end
 
 utilities.breadcrumbs = function ()
   local breadcrumbs = {}
@@ -568,7 +607,8 @@ utilities.breadcrumbs = function ()
   end
 
   function breadcrumbs:parent (count)
-    return self[#self-(count or 1)]
+    -- Note LuaJIT does not support __len, so this has to work even when that metamethod doesn't fire...
+    return self[#SILE.traceStack-(count or 1)]
   end
 
   function breadcrumbs:contains (needle)
@@ -581,4 +621,8 @@ utilities.breadcrumbs = function ()
   return breadcrumbs
 end
 
+utilities.formatNumber = require("core.utilities-numbers")
+
+utilities.collatedSort = require("core.utilities-sorting")
+
 return utilities
